/**
 * An abstract class, providing utitlity methods helpful when building controls
 *
 * @module Controls
 * @since 4.2
 * @class Control
 * @namespace AJS
 */
AJS.Control = Class.extend({

     INVALID: "INVALID",

    /**
     * An error for people trying to access private properties
     *
     * @method _throwReadOnlyError
     * @param property - property attempted to be read
     */
    _throwReadOnlyError: function (property) {
        new Error(this.CLASS_SIGNATURE + ": Sorry [" + property + "] is a read-only property");
    },

    /**
     * Allows binding of multiple events via a group. Event groups are stored under the _events property of the class.
     *
     * @method _assignEvents
     * @protected
     * @param {String} group - name of object group containing events
     * @param {String | HTMLElement | jQuery} target - element to bind events to
     */
    _assignEvents: function (group, target) {

        var key, handlers;
        var instance = this;

        this._activeEvents = this._activeEvents || {};

        if (typeof target === "string") {

            key = group + "/" + target;

            if (this._activeEvents[key]) {
                // We've already bound these events. No action needed.
                return;
            }

            handlers = this._activeEvents[key] = {};

            AJS.$.each(this._events[group], function(eventType, handler) {
                handlers[eventType] = function(event) {
                    handler.call(instance, event, AJS.$(this));
                };
                AJS.$(document).delegate(target, eventType, handlers[eventType]);
            });

        } else {

            target = AJS.$(target);

            if (target.length === 0) {
                // Warn: No target supplied.
                return;
            }

            if (this._activeEvents[group + "/" + target[0][AJS.$.expando]]) {
                // We've already bound these events. No action needed.
                return;
            }

            handlers = {};

            AJS.$.each(this._events[group], function(eventType, handler) {
                handlers[eventType] = function(event) {
                    handler.call(instance, event, AJS.$(this));
                };
                target.bind(eventType, handlers[eventType]);
            });

            key = group + "/" + (target[0] === window ? "window" : target[0][AJS.$.expando]);

            this._activeEvents[key] = handlers;
        }
    },

    /**
     * Allows unbinding of multiple events via a group. Event groups are stored under the _events property of the class.
     *
     * @method _assignEvents
     * @protected
     * @param {String} group - name of object group containing events
     * @param {String | HTMLElement | jQuery} target - element to bind events to
     */
    _unassignEvents: function (group, target) {

        var key, handlers, eventType;

        this._activeEvents = this._activeEvents || {};

        if (typeof target === "string") {

            key = group + "/" + target;
            handlers = this._activeEvents[key];

            if (!handlers) {
                // These events aren't bound on target. No action needed.
                return;
            }

            for (eventType in handlers) {
                AJS.$(document).undelegate(target, eventType, handlers[eventType]);
            }

        } else {

            target = AJS.$(target);

            if (target.length === 0) {
                // Warn: No target supplied.
                return;
            }

            key = group + "/" + target[0][AJS.$.expando];
            handlers = this._activeEvents[key];

            if (!handlers) {
                // These events aren't bound on target. No action needed.
                return;
            }

            // Note: You'd think we could just unbind these event handlers, but jQuery bug #7043 says no,
            // so we need to do the following ridiculous workaround for the time being.
            try {
                for (eventType in handlers) {
                    target.unbind(eventType, handlers[eventType]);
                }
            } catch (error) {
                var events = AJS.$.data(target[0], "events");
                if (events) {
                    for (eventType in handlers) {
                        var $handlers = events[eventType];
                        if (!$handlers) {
                            continue;
                        }
                        var i = $handlers.length;
                        while (i--) {
                            if ($handlers[i].handler === handlers[eventType]) {
                                $handlers.splice(i, 1);
                                if ($handlers.length === 0) {
                                    // Clean up when no events remain.
                                    delete events[eventType];
                                }
                                break;
                            }
                        }
                    }
                }
            }
        }

        delete this._activeEvents[key];
    },

    /**
     * @method _isValidInput
     * @return {Boolean}
     */
    _isValidInput: function () {
        return true;  
    },

    /**
     * A more descriptive way to handle key events. Use this method to delegate key events to the keys property map.
     *
     * An example of a keys property map looks like this (property names should be a shortcut as supported by
     * jira.keyboard.shortcutEntered; the 'onEdit' key handles editing the input):
     *
     * keys: {
     *      "ctrl+a": function (e) {
     *           // handle ctrl+a
     *      },
     *      "return": function (e) {
     *          // do something on enter/return
     *      },
     *      onEdit: function (e, character) {
     *         // handle input edited (character may be undefined, e.g., during backspace or delete).
     *      }
     * }
     *
     * @method _handleKeyEvent
     * @param {Object} e - event object
     */
    _handleKeyEvent: function (e) {
        var instance = this;
        if (instance._isValidInput(e)) {
            var SpecialKey = jira.keyboard.SpecialKey,
                shortcut = jira.keyboard.shortcutEntered(e);
            if (shortcut) {
                if (instance.keys[shortcut]) {
                    instance.keys[shortcut].call(instance, e);
                    return;
                } else if ((shortcut === SpecialKey.BACKSPACE || shortcut === SpecialKey.DELETE) && instance.keys.onEdit) {
                    instance.keys.onEdit.call(instance, e);
                    return;
                }
            }

            var character = jira.keyboard.characterEntered(e);
            if (character && instance.keys.onEdit) {
                instance.keys.onEdit.call(instance, e, character);
            }
        }
    },

    /**
     * Appends the class signature to the event name for more descriptive and unique event names.
     *
     * @method getCustomEventName
     * @param {String} methodName
     * @return {String}
     */
    getCustomEventName: function (methodName) {
        return (this.CLASS_SIGNATURE || "") + "_" + methodName;
    },

    /**
     * Gets default arguments to be passed to the custom event handlers
     *
     * @method _getCustomEventArgs
     * @protected
     * @return {Array}
     */
    _getCustomEventArgs: function () {
        return [this];
    },

    /**
     * Used to fire custom events on the AJS.Control instance.
     *
     * @method trigger
     * @param {String} event -- The name of the event to trigger.
     */
    trigger: function(event) {
        event = new jQuery.Event(event);
        AJS.$(this).trigger(event);
        return !event.isDefaultPrevented();
    },

    /**
     * Does the browser support css3 box shadows
     *
     * @method _supportsBoxShadow
     * @return {Boolean}
     */
    _supportsBoxShadow: function () {
        var s=document.body.style;
        return s.WebkitBoxShadow!== undefined||s.MozBoxShadow!==undefined||s.boxShadow!==undefined;
    },


    /**
     * Overrides default options with user options. If the element property is set to a field set, it will attempt
     * to parse options the options from fieldset
     *
     * @method _setOptions
     * @param options
     * @return {String | undefined} if invalid will return this.INVALID
     */
    _setOptions: function (options) {

        var element, optionsFromDOM;

        options = options || {};

        // just supplied element selector
        if (options instanceof AJS.$ || typeof options === "string") {
            options = {element: options};
        }

        element = AJS.$(options.element);

        optionsFromDOM = element.getOptionsFromAttributes();

        this.options = AJS.$.extend(true, this._getDefaultOptions(options), optionsFromDOM, options);

        if (element.length === 0) {
            return this.INVALID;
        }
    },

    /**
     * Gets position of carot in field
     *
     * @method getCaret
     * @param {HTMLElement} node
     * @return {Number} - The caret position within node, or -1 if some text is selected (and no unique caret position exists).
     */
    getCaret: function (node) {
        var startIndex = node.selectionStart;

        if (startIndex >= 0) {
            return (node.selectionEnd > startIndex) ? -1 : startIndex;
        }

        if (document.selection) {
            var textRange1 = document.selection.createRange();

            if (textRange1.text.length === 0) {
                var textRange2 = textRange1.duplicate();

                textRange2.moveToElementText(node); // Set textRange2 to select all text in node.
                textRange2.setEndPoint("EndToStart", textRange1); // Set the end point of textRange2 to the start point of textRange1.

                return textRange2.text.length;
            }
        }

        return -1;
    },


    /**
     * Delegates DOM rendering
     *
     * @method _render
     * @protected
     * @return {jQuery}
     */
    _render: function () {

        var i,
            name = arguments[0],
            args = [];

        for (i=1; i < arguments.length; i++) {
            args.push(arguments[i]);
        }

        return this._renders[name].apply(this, args);
    }
});
