/**
 * Creates dropdown menu functionality. It is <strong>STRONGLY</strong> advised that you create these objects through
 * the factory method @see AJS.DropDown.create
 *
 * @class DropDown
 * @namespace AJS
 * @extends AJS.Control
 */
AJS.DropDown = AJS.Control.extend({

    CLASS_SIGNATURE: "AJS_DROPDOWN",

    /**
     * @constructor
     * @param {Object | AJS.DropDown.Descriptor} options - @see AJS.DropDown.Descriptor
     */
    init: function (options) {

        var instance = this;

        if (!(options instanceof AJS.DropDown.Descriptor)) {
            this.options = new AJS.DropDown.Descriptor(options);
        } else {
            this.options = options;
        }
        
        this.layerController = new AJS.InlineLayer(this.options.allProperties());
        this.listController = this.options.listController();

        // we need to do cleanup if the inlinelayer is hidden by one of its own events
        this.layerController.onhide(function () {
           instance.hide();
        });

        this.layerController.contentChange(function () {
            instance.listController.removeAllItems();
            instance.layerController.layer().find("li").each(function () {
                instance.listController.addItem(new AJS.ListItem({ element: this }));
            });
            instance.listController.shiftFocus(0);
        });

        this.trigger(this.options.trigger()); // bind trigger events

        this._applyIdToLayer();
    },

    /**
     * Shows dropdown, in the case of an ajax dropdown this will make a request to get content if there isn't already some
     *
     * @method show
     */
    show: function () {
        this.trigger().addClass(AJS.ACTIVE_CLASS);
        this.layerController.show();
        this.listController.shiftFocus(0);
    },

    /**
     * Hides dropdown
     *
     * @method hide
     */
    hide: function () {
        this.trigger().removeClass(AJS.ACTIVE_CLASS);
        this.layerController.hide();
        this.listController.trigger("blur");
    },

    /**
     * Hides and shows dropdown
     *
     * @method toggle
     */
    toggle: function () {
        if (this.layerController.isVisible()) {
            this.hide();
        } else {
            this.show();
        }
    },

    /**
     * Sets/Gets content. Delegates to layer controller.
     *
     * @method trigger
     * @param {jQuery} content
     * @return {jQuery}
     */
    content: function (content) {
        if (content) {
            this.layerController.content(content);
        } else {
            return this.layerController.content();
        }
    },

    /**
     * Sets/Gets trigger. If setting, unbinds events of previous trigger (if there was one), binding events to new one.
     *
     * @method trigger
     * @param {jQuery} trigger
     * @return {jQuery}
     */
    trigger: function (trigger) {
        if (trigger) {

            if (this.options.trigger()) {
                this._unassignEvents("trigger", this.options.trigger());
            }

            this.options.trigger(AJS.$(trigger));

            this.layerController.offsetTarget(this.options.trigger());

            this._assignEvents("trigger", this.options.trigger());
            
        } else {
            return this.options.trigger();
        }
    },

    _applyIdToLayer: function () {
        if (this.trigger().attr("id")) {
            this.layerController.layer().attr("id", this.trigger().attr("id") + "_drop");
        }
    },

    _events: {
        trigger: {
            click: function (e) {
                this.toggle();
                e.preventDefault(); // in-case we are a link
            }
        }
    }

});

// static
AJS.$.extend(AJS.DropDown, {
    TRIGGER_SELECTOR: ".aui-dropdown-trigger",
    CONTENT_SELECTOR: ".aui-dropdown-content"
});
