/**
 * A multiselect list that can be queried and suggestions selected via a dropdown. Suggestions are retrieved via AJAX.
 *
 * @module Controls
 * @since 4.2
 * @namespace AJS
 * @class MultiSelect
 * @extends AJS.QueryableSelect
 * @requires AJS.$, AJS.SelectModel, AJS.QueryableSelect, AJS.Dropdown, AJS.List, AJS.Control, AJS.Group, AJS.Lozenge
 */

AJS.MultiSelect = AJS.QueryableSelect.extend({

    /**
     * This constructor:
     * <ul>
     *  <li>Overrides default options with user options</li>
     *  <li>Inserts an input field before dropdown</li>
     *  <li>Adds items currently selected in the &lt;select&gt; as items sitting on the top of the textarea</li>
     * <ul>
     *
     * @constructor
     * @param options
     */
    init: function (options) {

        if (this._super(options) === this.INVALID) {
            return this.INVALID;
        }

        var instance = this;

        // sets max min withs of field
        this._setTextareaDimThresholds();

        this.lozengeGroup = new AJS.LozengeGroup();
        this._assignEvents("lozengeGroup", this.lozengeGroup);

        this._restoreSelectedOptions();

        return this;
    },

    /**
     * Gets default options
     *
     * @method _getDefaultOptions
     * @protected
     * @return {Object}
     */
    _getDefaultOptions: function () {
        return AJS.$.extend(true, this._super(), {
            minRoomForText : 50,
            errorMessage: AJS.params.multiselectGenericError,
            ajaxOptions: {
                minQueryLength: 1
            },
            showDropdownButton: true
        });
    },

    /**
     * Appends furniture around specified dropdown element. This includes:
     *
     * <ul>
     *  <li>errorMessage - A container for error messages is created but not appended until needed</li>
     *  <li>selectedItemsWrapper - A wrapper for selected items</li>
     *  <li>selectedItemsContainer - A container for selected items</li>
     * </ul>
     *
     * @method _createFurniture
     * @protected
     */
    _createFurniture: function (disabled) {

        if (this.model.$element.prev().hasClass("ajs-multi-select-placeholder")) {
            this.model.$element.prev().remove();
        }

        this._super(disabled);

        // remove placeholder if there is one. This placeholder, takes up the space that the multi-select control will
        // while the page is being loaded and the "real" control has not been inserted. i.e Stops the page jumping around.

        if (!disabled) {
            this.$errorMessage = this._render("errorMessage");
            this.$selectedItemsWrapper = this._render("selectedItemsWrapper")
                    .appendTo(this.$container);
            this.$selectedItemsContainer = this._render("selectedItemsContainer")
                    .appendTo(this.$selectedItemsWrapper);
            }
    },

    /**
     * Assigns events furniture
     *
     * @method _assignEventsToFurniture
     */
    _assignEventsToFurniture: function () {
        this._super();
        this._assignEvents("body", document);
        this._assignEvents("selectedItemsContainer", this.$selectedItemsContainer);
    },

    /**
     * Unselects item in model and removes item from selectedItemsContainer
     *
     * @method removeItem
     * @param {Object} descriptor
     */
    removeItem: function (descriptor) {
        this.model.setUnSelected(descriptor);
        this.updateItemsIndent();
     }, 

    /**
     * Adds items currently selected in the &lt;select&gt; as items sitting on the top of the textarea
     *
     * @method _restoreSelectedOptions
     */
    _restoreSelectedOptions: function () {
        var instance = this;

        // creates selected "button" style representation
        AJS.$.each(this.model.getSelectedDescriptors(), function () {
            instance._addItem(this);
        });
        this.updateItemsIndent();
    },

    /**
     * Is true if:
     * <ol>
     *      <li>The text selection is at the start or there is no value in field</li>
     *      <li>Lozenge group has not already been enabled.</li>
     *      <li>There are items in the lozenge group</li>
     * </ol>
     *
     * @method _shouldEnableLozengeGroup
     * @return {Boolean}
     */
    _shouldEnableLozengeGroup: function () {
        return this.lozengeGroup.items.length > 0 && this.lozengeGroup.index < 0 && (this.$field.val().length === 0 || this.getCaret(this.$field[0]) === 0);
    },

    /**
     * Handling of backspace in textarea. If there is no characters will select the last item in the selectedItemsContainer.
     */
    _handleBackSpace: function () {
        var instance = this;
        if (this._shouldEnableLozengeGroup()) {
            setTimeout(function() {
                instance.lozengeGroup.shiftFocus(-1);
            }, 0);
        } else {
            this.$field.one("keyup", function() {
                instance._handleCharacterInput();
            });
        }
    },

     /**
     * Handling of delete in textarea.
     */
    _handleDelete: function () {
        if  (AJS.$.trim(this.$field.val()) !== "") {
            // delay until character is actually removed from textarea
            var instance = this;
            this.$field.one("keyup", function () {
                instance._handleCharacterInput();
            });
        }
    },

    /**
     * Handling of left key in textarea. If there is no characters will select the last item in the selectedItemsContainer.
     */
    _handleLeft: function () {
        if (this._shouldEnableLozengeGroup()) {
            var instance = this;
            setTimeout(function() {
                instance.lozengeGroup.shiftFocus(-1);
            }, 0);
        }
    },

    /**
     * Handle text paste event.
     */
    _handlePaste: function() {
        // Remove whitespace such as newlines, tabs, etc.
        this.$field.val(AJS.$.trim(this.$field.val()).replace(/\s+/g, " "));
        this._handleCharacterInput();
    },

    /**
     * Sets the min and max widths of the textarea. As the user selects items the textarea can grow from the min width
     * to max width.
     *
     * @method _setTextareaDimThresholds
     */
    _setTextareaDimThresholds: function () {
        this.maxWidth = this.options.maxWidth || this.$container.attr("scrollWidth");
        this.minWidth = this.options.minWidth || this.$container.attr("scrollWidth");
        if (this.minWidth > this.maxWidth) {
            this.minWidth = this.maxWidth;
        }
    },

    /**
     * Updates the padding left and padding top based on the area occupied by the selected items
     *
     * @method updateItemsIndent
     */
    updateItemsIndent: function () {

        var inputIndent = this._getInputIndent();

        // First set $field's new height, which may trigger vertical scrollbars when
        // in a dialog ...
        this.$field.css({
            paddingTop: inputIndent.top,
            paddingLeft: inputIndent.left,
            width: 0
        });
        // ... *then* set $field's width based on the width of $container, which will
        // have shrunken after the scrollbars appeared.
        this.$field.css({
            width: this.$container.width() - inputIndent.left - 21
        });

        if (this.currentTopOffset && this.currentTopOffset !== inputIndent.top) {
            this.$container.trigger("multiSelectHeightUpdated", [this]);
        }

        // otherwise ie does not update indent
        if (AJS.$.browser.msie) {
            this.$field.val(this.$field.val() + " ");
            this.$field.val(this.$field.val().replace(/\s$/, ""));
        }

        this.currentTopOffset = inputIndent.top;
    },

    /**
     * Checks if the item has already been selected/added
     *
     * @method _isItemPresent
     * @param descriptor - JSON describing suggestion/option
     * @return {Boolean}
     */
    _isItemPresent: function (descriptor) {
        var duplicate = false;
        var label = descriptor.label();
        AJS.$.each(this.lozengeGroup.items, function () {
            if (this.options.label === label) {
                duplicate = true;
                return false; // bail
            }
        });
        return duplicate;
    },

    /**
     * Adds selected suggestion to the selected items, and marks it as selected in the model.
     *
     * @method _addItem
     * @param {Object} descriptor - JSON describing suggestion/option
     */
    _addItem: function(descriptor) {

        // this descriptor is for the lozenge so we don't want to use the same descriptor but a copy instead. We don't
        // want our descriptor properties to change through a reference we were unaware of.
        if (descriptor instanceof AJS.ItemDescriptor) {
            descriptor = AJS.copyObject(descriptor.allProperties(), false);
        }

        descriptor.value = AJS.$.trim(descriptor.value);
        descriptor.label = AJS.$.trim(descriptor[this.options.itemAttrDisplayed]) || descriptor.value;
        descriptor.title = AJS.$.trim(descriptor.title) || descriptor.label;

        descriptor = new AJS.ItemDescriptor(descriptor);
        
        if (this._isItemPresent(descriptor)) {
            return;
        }

        var lozenge = new AJS.Lozenge({
            label: descriptor.label(),
            title: descriptor.title(),
            container: this.$selectedItemsContainer
        });

        this.lozengeGroup.addItem(lozenge);
        this._assignEvents("lozenge", lozenge);

        this.model.setSelected(descriptor);
        this.updateItemsIndent();

        this.dropdownController.setPosition(); // update position incase another line has been added
    },

    /**
     * Adds multiple items
     *
     * @method _addMultipleItems
     * @param {Array} items - Array of item descriptors. e.g {value: "The val", label: "Label to be displayed in suggestions"}
     * @param {Boolean} removeOnUnSelect - If set to true, if the item is removed from the control (unselected), the option
     * will also be deleted from the select model also. This means it will not appear in the suggestions dropdown.
     */
    _addMultipleItems: function(items, removeOnUnSelect) {

        var instance = this;

        AJS.$.each(items, function (i, descriptor) {
            if (removeOnUnSelect) {
                descriptor.removeOnUnSelect = true;
            }
            instance._addItem(descriptor);
        });
    },

    /**
     * Calculates the width taken by selected items
     *
     * @method _getTargetItemContainerWidth
     * @return {Number} width
     */
    _getTargetItemContainerWidth: function () {

        var lozenges = this.lozengeGroup.items,
            width = parseInt(this.$selectedItemsContainer.css("paddingLeft"), 10) +
                    parseInt(this.$selectedItemsContainer.css("paddingRight"), 10);

        for (var i = lozenges.length-1; i >= 0; i--) {
            var $item = lozenges[i].$lozenge;
            width = width
                  + $item.outerWidth()
                  + parseInt($item.css("marginLeft"), 10)
                  + parseInt($item.css("marginRight"), 10);
        }
        return width;
    },

    /**
     * Determines correct top and left indent for textarea, based on the area taken by selected items
     *
     * @method _getInputIndent
     * @return {Object}
     */
    _getInputIndent: function () {

        // TODO incorporate min/max width options

        var top,
            left,
            iconArea = 22, // Icon width of 16px and 4px on the right and then 2px margin on the left of the items container
            paddingLeft = 2,
            paddingTop = 4,
            indent = { top: paddingTop, left: paddingLeft },
            lastLozengeIndex = this.lozengeGroup.items.length - 1,
            $last;

        if (lastLozengeIndex >= 0) {
            $last = this.lozengeGroup.items[lastLozengeIndex].$lozenge;
            top = $last.attr("offsetTop");
            left = $last.attr("offsetLeft") + $last.outerWidth();
            if (left > this.$container.width() - iconArea - this.options.minRoomForText) {
                top += $last.attr("offsetHeight");
                left = 0;
            }

            indent.top += top;
            indent.left += left;
        }

        return indent;
    },

    /**
     * @method _selectionHandler - Handle when a suggestion is accepted
     * @param selected
     * @param e
     */
    _selectionHandler: function (selected, e) {

        var instance = this;
        
        selected.each(function () {
            instance._addItem(AJS.$.data(this, "descriptor"));
        });

        this.$field.val("").focus().scrollIntoView({ margin: 20 });
        this.hideSuggestions();
        this.hideErrorMessage();
        this.model.$element.trigger("change");
        
        e.preventDefault();
    },

    /**
     * @method isValidItem - Determines whether the given value represents a valid item
     * @param {String} itemValue
     * @return {Boolean}
     */
    isValidItem: function(itemValue) {
        var suggestedItemDescriptor = this.listController.getFocused().data("descriptor");

        if (!suggestedItemDescriptor) {
            return false;
        }

        itemValue = itemValue.toLowerCase();

        return itemValue === AJS.$.trim(suggestedItemDescriptor.label.toLowerCase()) ||
               itemValue === AJS.$.trim(suggestedItemDescriptor.value.toLowerCase());
    },

    /**
     * @method showErrorMessage - Show an error message near this field
     * @param {String} value (optional) - The user input text responsible for the error
     */
    showErrorMessage: function(value) {
        this.$errorMessage.text(AJS.format(this.options.errorMessage, value || this.$field.val()));
        var $container = this.$container.parent(".field-group");
        if ($container.length === 0){
            // not in aui 
            $container = this.$container.parent(".frother-control-renderer");
            this.$errorMessage.prependTo($container);
        } else {
            this.$errorMessage.appendTo($container);
        }
    },

    /**
     * @method hideErrorMessage - Hide the error message
     */
    hideErrorMessage: function() {
        this.$errorMessage.remove();
    },

    /**
     * @method handleFreeInput - Handle the case where text remains "unlozenged" in the text field
     */
    handleFreeInput: function() {
        var value = AJS.$.trim(this.$field.val()),
            descriptor;

        if (value) {
            descriptor = this.model.getDescriptor(value);
            if (descriptor) {
                this._addItem(descriptor);
                this.model.$element.trigger("change");
            } else {
                this.showErrorMessage(value);
                return;
            }
        }
        this.hideErrorMessage();
        this.$field.val("");
    },

    /**
     * Submits form
     * @method submitForm
     */
    submitForm: function () {
        if (this.$field.val().length === 0 && !this.suggestionsVisible) {
            AJS.$(this.$field[0].form).submit(); // submit on enter if field is empty
        }
    },


    keys: {

        left: function () {
            this._handleLeft();
        },

        backspace: function () {
            this._handleBackSpace();
        },

        del: function () {
            this._handleDelete();
        },
        
        "return": function (e) {
            this.submitForm();
            e.preventDefault();
        }
    },

    _events: {
        body: {
            // handling for the case where control is in a tab, and as a result hidden.
            tabSelect: function () {
                if (this.$field.is(":visible")) {
                    this._setTextareaDimThresholds();
                    this.updateItemsIndent();
                }
            },
            // Bulk operations don't yet use AUI forms.
            bulkTabSelect: function() {
                if (this.$field.is(":visible")) {
                    this._setTextareaDimThresholds();
                    this.updateItemsIndent();
                }
            }
        },
        field: {
            blur: function() {
                this.handleFreeInput();
                this.lozengeGroup.trigger("blur");
                this.hideSuggestions();
            },
            paste: function() {
                // This must be delayed because the pasted text has not been added yet.
                setTimeout(AJS.$.proxy(this, "_handlePaste"), 0);
            },
            "keydown keypress": function(event) {
                if (this.lozengeGroup.index >= 0) {
                    if (jira.keyboard.SpecialKey.fromKeyCode(event.keyCode) in this.lozengeGroup.keys) {
                        event.preventDefault();
                    } else if (jira.keyboard.SpecialKey.fromKeyCode(event.keyCode) === "return") {
                        this.submitForm();
                        event.preventDefault();
                    } else {
                        var instance = this;
                        this.$field.one("keyup", function() {
                            instance._handleCharacterInput();
                        });
                        this.lozengeGroup.trigger("blur");
                    }
                }
            },
            click: function() {
                this.lozengeGroup.trigger("blur");
                this.$field.focus();
            }
        },
        lozengeGroup: {
            focus: function() {
                this.$field.focus();
                this.hideSuggestions();
                this._unassignEvents("keys", this.$field);
            },
            blur: function() {
                this._assignEvents("keys", this.$field);
                if (this.$field.val()) {
                    this._handleCharacterInput();
                }
            }
        },
        lozenge: {
            remove: function(event) {
                this.removeItem(this.model.getDescriptor(event.target.options.label));
            }
        },
        selectedItemsContainer: {
            click: function(event) {
                // Ignore clicks not directly on this.$selectedItemsContainer.
                if (event.target === event.currentTarget) {
                    this.lozengeGroup.trigger("blur");
                    this.$field.focus();
                }
            }
        }
    },

    _renders: {
        errorMessage: function () {
            return AJS.$('<div class="error" />');
        },
        selectedItemsWrapper: function () {
            return AJS.$('<div class="representation"></div>');
        },
        selectedItemsContainer: function () {
            return AJS.$('<ul class="items" />');
        }
    }
});
