/**
 * A select list that can be queried and suggestions selected via a dropdown. Suggestions are retrieved via AJAX.
 *
 * todo: This is currently an abstract object for multiselect, however should be fixed to represent single select.
 *
 * @module Controls
 * @since 4.2
 * @namespace AJS
 * @class QueryableSelect
 * @extends AJS.QueryableDropdown
 * @requires AJS.$, AJS.SelectModel, AJS.Dropdown, AJS.List, AJS.Control
 */


AJS.QueryableSelect = AJS.QueryableDropdown.extend({

    /**
     * Overrides default options with user options. Inserts an input field before dropdown.
     * 
     * @constructor
     * @param options
     */
    init: function (options) {

        var instance = this;

        if (this._setOptions(options) === this.INVALID) {
            return this.INVALID;
        }

        this.options.element.hide();

        if (this.options.disabled) {
            this._createFurniture(true);
            return this;
        }

        this.model = new AJS.SelectModel({
            element: this.options.element,
            removeOnUnSelect: this.options.removeOnUnSelect
        });

        // Add the visual representation
        this._createFurniture();

        this.dropdownController = AJS.InlineLayer.create({
            alignment:AJS.LEFT,
            offsetTarget: this.$field,
            content: AJS.$(".aui-list", this.$container)
        });

        this.listController = new AJS.List({
            containerSelector: AJS.$(".aui-list", this.$container),
            groupSelector: "ul.aui-list-section",
            itemSelector: "li",
            matchingStrategy: this.options.matchingStrategy,
            selectionHandler: function (e) {
                instance._selectionHandler(this.getFocused(), e);
                return false;
            }
        });

        this._assignEventsToFurniture();
    },
    
    /**
     * Appends furniture. This includes:
     *
     * <ul>
     *  <li>Field - text field used fro querying</li>
     *  <li>Container - Wrapper used to contain all furniture</li>
     *  <li>Dropdown Icon - Button in right of field used to open dropdown via mouse</li>
     *  <li>Suggestions container - Dropdown that contains suggestions</li>
     * </ul>
     * 
     * @method _createFurniture
     * @param {Boolean} disabled - If disabled will render a standard text field
     */
    _createFurniture: function (disabled) {

        var id = this.model.$element.attr("id");

        if (disabled) {
            this.model.$element.replaceWith(this._render("disableSelectField", id));
        } else {
            this.$container = this._render("container", id);
            this.$field = this._render("field", id).appendTo(this.$container);
            this.$container.append(this._render("suggestionsContainer", id));
            this.$container.insertBefore(this.model.$element);
            this.$dropDownIcon = this._render("dropdownAndLoadingIcon", this._hasDropdownButton()).appendTo(this.$container);
        }
    },

    /**
     * Gets value for suggestion/option mirroring user input
     *
     * @method _getUserInputValue
     * @protected
     * @return {String}
     */
    _getUserInputValue: function () {
        return this.options.uppercaseUserEnteredOnSelect ? this.$field.val().toUpperCase(): this.$field.val();
    },

    /**
     * Adds the user's input as an option in the suggestions and the select model
     *
     * @method _handleUserInputOption
     */
    _handleUserInputOption: function () {

        var groupDescriptor;

        if (!this.hasUserInputtedOption() || this.$field.val().length === 0) {
            return;
        }

        groupDescriptor = new AJS.GroupDescriptor({
            type: "optgroup",
            label: "user inputted option",
            weight: 9999,
            showLabel: false,
            replace: true
        });

        groupDescriptor.addItem(new AJS.ItemDescriptor({
            value: this._getUserInputValue(),
            label: this.$field.val(),
            labelSuffix: " (" + this.options.userEnteredOptionsMsg + ")",
            title: this.$field.val(),
            allowDuplicate: false
        }));

        this.model.appendOptionsFromJSON([groupDescriptor]);
    },

    /**
     * Returns weather or not an option will be appended to the bottom of the suggestions, mirroring the user input.
     * The will be true if at invocation the userEnteredOptionMsg is set in the options object. The value of this property
     * will be displayed in brackets after the suggestion.
     *
     * @method hasUserInputtedOption
     * @return {Boolean}
     */
    hasUserInputtedOption: function () {
        return this.options.userEnteredOptionsMsg;      
    },

    /**
     * Some special handling for user inputted option, and removal of item in select list if we are replacing them from
     * the server.
     *
     * @method _handleSuggestionResponse
     * @override
     * @protected
     * @param {Array} data
     */
    _handleSuggestionResponse: function (data) {

        // If we are querying the server then the server will return the full result set to be displayed. We do not
        // want any linguring options in there.
        if (this.options.ajaxOptions.query) {
            this.model.clearUnSelected();
        }

        // we have to do this again as we clear all the unselected options above, which includes the user inputted option
        this._handleUserInputOption();

        this._super(data);
    },

    /**
     * Adds supplied suggestions to dropdown and &lt;select&gt; list
     *
     * @method _setSuggestions
     * @param {Object} data - JSON representing suggestions
     */
    _setSuggestions: function (data) {
        if (data) {
            this.model.appendOptionsFromJSON(data);
            this._super(this.model.getUnSelectedDescriptors());
        } else {
            this.hideSuggestions();
        }
    },

    _renders: {
        field: function (idPrefix) {
             //  the wrap="off" attribute prevents text from growing under the labels. It doesn't prevent linebreaks
            return AJS.$('<textarea autocomplete="off" id="' + idPrefix + '-textarea" class="aui-field" wrap="off"></textarea>');
        },
        disableSelectField: function (id) {
            return AJS.$("<input type='text' class='long-field' name='" + id + "' id='" + id + "' />");
        },
        container : function (idPrefix) {
            return AJS.$('<div class="multi-select" id="' + idPrefix +'-multi-select">');
        },
        suggestionsContainer : function (idPrefix) {
            return AJS.$('<div class="aui-list" id="' + idPrefix + '-suggestions" tabindex="-1"></div>');
        }
    }
});
