/**
 * Converts a standard &lt;select&gt; field into a version picker
 *
 * HTML:
 *
 * <select size="5" name="fixVersions" multiple="multiple" id="fixVersions">
 *   <optgroup label="Unreleased Versions">
 *       <option value="10010">Scott Harwood</option>
 *       <option value="10011">Scott Hughes</option>
 *       <option value="10012">Scott Farquar</option>
 *   </optgroup>
 *
 * JavaScript:
 * 
 * jira.app.findVersionSelectAndConvertToPicker(document.body, "#fixVersions")
 *
 * @method findVersionSelectAndConvertToPicker
 * @param {String, HTMLElement, JQuery} context - context to perform the selector search. If none supplied will search entire DOM
 * @param {String} selector - jQuery selector matching either a parent of the &lt;select&gt; or the select itself.
 *
 * If none supplied will used default selector. Which is:
 *
 * <pre>
 *  div.aui-field-versionspicker, // aui forms
 *  td.aui-field-versionspicker, // convert to subtask and move
 *  tr.aui-field-versionspicker // bulk edit
 * </pre>
 *
 * @namespace jira.app
 */
jira.app.findVersionSelectAndConvertToPicker = function () {

    function createPicker($selectField) {
        new AJS.MultiSelect({
           element: $selectField,
           itemAttrDisplayed: "label",
           errorMessage: AJS.params.multiselectVersionsError
        });
    }

    function locateSelect(parent) {

        var $parent = AJS.$(parent),
            $selectField;

        if ($parent.is("select")) {
            $selectField = $parent;
        } else {
            $selectField = $parent.find("select");
        }

        return $selectField;
    }

    var DEFAULT_SELECTORS = [
        "div.aui-field-versionspicker.frother-control-renderer", // aui forms
        "td.aui-field-versionspicker.frother-control-renderer", // convert to subtask and move
        "tr.aui-field-versionspicker.frother-control-renderer" // bulk edit
    ];

    return function (context, selector) {
        
        selector = selector || DEFAULT_SELECTORS.join(", ");

        AJS.$(selector, context).each(function () {

            var $selectField = locateSelect(this);

            if ($selectField.length) {
                createPicker($selectField);
            } else {
                /* [logging] */
                AJS.log("jira.app.convertToVersionPicker: Tried to create version picker using [" + selector + "] "
                        + "however selector does not match or could not find <select>");
                /* [logging] end */
            }

        });
        
    };
}();
